function [Uout,eigvals] = zoloeig(H,normH)
%ZOLO-EIG    Eigendecomposition of symmetric A via Zolotarev-approximation,
%  uses Zolopd for polar decomposition
%   [U,D] = ZOLOEIG(A) computes the eigenvalues (the diagonal elements
%   of of D) and an orthogonal matrix of eigenvectors (the columns of V)
%   of the symmetric matrix A. This function makes use of the function
%   QDWH that implements the QR-based dynamically weighted Halley
%   iteration for the polar decomposition.
%   [U,D] = ZOLOEIG(A,normA,minlen,a,b,shift) includes the optional
%   input arguments
%     normA: norm(A,'fro'), which is used in the recursive calls.  
%     minlen: the matrix size at which to stop the recursions (default 1).
%     a: an estimate of norm(A,2).
%     b: an estimate of min(svd(A)).
%     shift: a scalar such that H-shift*I has a near-equal number of
%            positive and negative eigenvalues.
  
if nargin < 2 || isempty(normH); normH = norm(H,'fro'); end

[Uout,eigvals] = zoloeigrep(H,normH);

Uout = 3/2*Uout-Uout*(Uout'*Uout)/2; % Newton-Schulz postprocessing.

if nargout == 1; Uout = diag(eigvals); end


function [Uout,eigvals] = zoloeigrep(H,normH,minlen,a,b,shift)
  
backtol = 1e-15; % tolerance for relative backward error
n = length(H);
if nargin < 2; normH = norm(H,'fro'); end

% If already almost diagonal, then return trivial solution.
if norm(H-diag(diag(H)),'fro')/normH < backtol;
    Uout = eye(n); eigvals = diag(H)'; timeest = 0; return
end

H = (H+H')/2;   % Needed for recursive calls, due to roundoff.

if nargin < 3; minlen = 1; end
if nargin < 6 || isempty(shift)
   % Determine shift: approximation to median(eig(H)).
   shift = median(diag(H));
end

% Estimate a,b.
if nargin < 4 || isempty(a); a = normest(H-shift*eye(n),3e-1); end
if nargin < 5 || isempty(b); b = .9/condest(H-shift*eye(n)); end

% Compute polar decomposition via Zolotarev-pd.
%U = qdwh(H-shift*eye(n),a,b); % for qdwh-eig
U = zolopd(H-shift*eye(n),a,b);

% Orthogonal projection matrix.
U = (U+eye(n))/2;

% Subspace iteration
[U1,U2] = subspaceit(U);
minoff = norm(U2'*H*U1,'fro')/normH; % backward error

if minoff > backtol
    % 'Second subspace iteration'   
    [U1,U2] = subspaceit(U,0,U1);
    minoff = norm(U2'*H*U1,'fro')/normH; % backward error
%    minoff
end

if minoff > backtol
    for irand = 1:2
        % Redo subspace iteration with randomization.
        [U1b,U2b] = subspaceit(U,1);
        minoff2 = norm(U2b'*H*U1b,'fro')/normH; % backward error
        if minoff > minoff2; U1 = U1b; U2 = U2b; end     % choose better solution
    end
end

    
% One step done; further blocks.

eigvals1 = []; eigvals2 = [];    
if length(U1(1,:)) > minlen
    [Ua eigvals1] = zoloeigrep(U1'*H*U1,normH,minlen);
     U1 = U1*Ua;
    %    clear Ua
elseif length(U1(1,:)) > 1 % use MATLAB to complete
    [Ua D] = eig(U1'*(H*U1)); U1 = U1*Ua;
    eigvals1 = diag(D)';
    %    clear Ua
end
if length(U2(1,:)) > minlen
    [Ua eigvals2] = zoloeigrep(U2'*H*U2,normH,minlen);
    U2 = U2*Ua;
    %    clear Ua
elseif length(U2(1,:)) > 1 % use MATLAB to complete
    [Ua D] = eig(U2'*(H*U2)); U2 = U2*Ua;
    eigvals2 = diag(D)';
end

Uout = [U1 U2];
% Collect eigvals
eigvals = [eigvals1 eigvals2];
if length(U1(1,:)) == 1
eigvals = [U1'*H*U1 eigvals];
end
if length(U2(1,:)) == 1
eigvals = [eigvals U2'*H*U2];    
end


