function [U,H,m,it] = zolopd(A,alpha,L)
% zolopd; polar decomposition via Zolotarev-approximation
% U is the computed polar factor of A
% m is the Zolotarev degree
% H is the computed Hermitian factor of A
% it is the iteration number, usually 2 
%
% alpha \approx ||A||
% L approx 1/cond(A)

[m,n] = size(A);

if m == n && norm(A-A','fro')/norm(A,'fro') < 1e-14;
   symm = 1;else   symm = 0;
end

%Estimate extremal singular values
if nargin < 2 || isempty(alpha) % estimate for largest singular value of A
   alpha = normest(A,0.1);
end

% Scale original matrix to form X0.
U = A/alpha; %Uprev = U;

if nargin < 3 || isempty(L) % estimate for smallest singular value of U.
   Y = U; if m > n, [Q,Y] = qr(U,0); end
   smin_est =  norm(Y,1)/condest(Y);  % Actually an upper bound for smin.
   L = smin_est/sqrt(n);   
end

U = U/L; % normalize ||U||=~1
con = 1/L;

it = 0;
m = choosem(con);
if con<2, 
    itmax = 1;
else
    itmax = 2; 
end

while it<itmax
    it = it+1;

kp = 1/con;
alpha = acos(kp);
K = mellipke(alpha); % modified elliptic functions
m_zol = m; c = zeros(1,2*m_zol);
for ii = 1:2*m_zol,
   [sn,cn,tn] = mellipj(ii*K/(2*m_zol+1),alpha);   
   c(ii) = sn^2/cn^2;
end

%determine rational function
ff = @(x) x;
for ii = 1:m
    ff = @(x) ff(x).*(x.^2+c(2*ii))./(x.^2+c(2*ii-1));    
end

%U = computeAA(U,c,it,'cholq');% possible option: choleskyQR twice
%U = computeAA(U,c,it,'invse');% not recommended: explicit inversion
U = computeAA(U,c,it);

U = U/ff(1);% normalize so that min(svd(A))=1

if con<2; con = max(ff(con)/ff(1),1);break;end % if k(A) small, one step is enough
con = max(ff(con)/ff(1),1);
if symm;    U=(U'+U)/2;end    % force symmetry if A symmetric

end % end of iteration

U = (3/2)*U-U*(U'*U)/2; % Newton-Schulz refinement.

%clear Upre Cinv Q
if nargout > 1
H = U'*A; 
H = (H+H')/2;
end

function m = choosem(con);  % choose the Zolotarev degree according to the condnum
    if con<1.001;m=2;        
        elseif con<=1.01; m = 3;            
        elseif con<=1.1; m = 4;            
        elseif con<=1.2; m = 5;            
        elseif con<=1.5; m = 6;
        elseif con<=2; m = 8;   % one-step convergence till here
    elseif con<6.5; m = 2;
    elseif con<180; m = 3;        
    elseif con<1.5*10^4; m = 4;
    elseif con<2*10^6; m = 5;
    elseif con<1*10^9; m = 6;
    elseif con< 3*10^12; m = 7;
    else m = 8;
    end
    
function AA = computeAA(A,c,it,howqr)
if nargin<4, howqr = 'house'; end
AA = A;
r = length(c)/2; [m,n] = size(A);
for ii=1:r
    enu = 1;
    for jj=1:r
        enu = enu*(c(2*ii-1)-c(2*jj));
    end
    den = 1;
    for jj = 1:r
        if ii~=jj
        den = den*(c(2*ii-1)-c(2*jj-1));
        end
    end
    
    if it<=1 && max(c(1:end-1))>1e2    
        if howqr == 'house' % QR-based
    [AAA,~]=qr([A;sqrt(c(2*ii-1))*eye(n)],0);   
    AA=AA-enu/den/sqrt(c(2*ii-1))*AAA(1:m,:)*AAA(m+1:end,:)' ;       
        else % Cholesky
    II = sqrt(c(2*ii-1))*eye(n);     
    R = A'*A + (c(2*ii-1))*eye(n);
    R = chol(R);
    Q = A/R; 
    II = II/R; 
    if condest(R)>10,  % repeat
    R = Q'*Q + II'*II;
    R = chol(R);
    Q = Q/R; 
    II = II/R; 
    Rall = R*Rall;
    end    
    AA = AA-enu/den/sqrt(c(2*ii-1))*(Q*II');
    end
    else        
    Cinv = chol((A'*A)+c(2*ii-1)*eye(n));
    Qtmp = A/Cinv;Qtmp=Qtmp/Cinv';
    AA = AA-enu/den*Qtmp ;   
    end            
end    


function [k,e] = mellipke(alpha,tol)
%ELLIPKE Complete elliptic integral. Modified from Matlab's built-in code
%for improved accuracy. 
if nargin<1
  error(message('MATLAB:ellipke:NotEnoughInputs')); 
end
m = sin(alpha) * sin(alpha);
m1 = cos(alpha) * cos(alpha);

%classin = superiorfloat(m);
classin='double';

if nargin<2, tol = eps(classin); end
if ~isreal(m),
    error(message('MATLAB:ellipke:ComplexInputs'))
end
if isempty(m), k = zeros(size(m),classin); e = k; return, end
if any(m(:) < 0) || any(m(:) > 1), 
  error(message('MATLAB:ellipke:MOutOfRange'));
end

a0 = 1;
b0 = cos(alpha);
s0 = m;
i1 = 0; mm = 1;
while mm > tol
    a1 = (a0+b0)/2;
    b1 = sqrt(a0.*b0);
    c1 = (a0-b0)/2;
    i1 = i1 + 1;
    w1 = 2^i1*c1.^2;
    mm = max(w1(:));
    s0 = s0 + w1;
    a0 = a1;
    b0 = b1;
end
k = pi./(2*a1);
e = k.*(1-s0/2);
% im = find(m ==1);
% if ~isempty(im)
%     e(im) = ones(length(im),1);
%    k(im) = inf;
% end


function [sn,cn,dn] = mellipj(u,alpha,tol)
%mELLIPJ Jacobi elliptic functions. MATLAB's built-in code, modified for improved accuracy. 
%   [SN,CN,DN] = ELLIPJ(U,M) returns the values of the Jacobi elliptic 
%   functions Sn, Cn and Dn, evaluated for corresponding elements of 
%   argument U and parameter M.  U and M must be arrays of the same 
%   size or either can be scalar.  As currently implemented, M is 
%   limited to 0 <= M <= 1. 
%
%   [SN,CN,DN] = ELLIPJ(U,M,TOL) computes the elliptic functions to
%   the accuracy TOL instead of the default TOL = EPS.  
%

if nargin<2
  error(message('MATLAB:ellipj:NotEnoughInputs')); 
end

u=real(u);
alpha=real(alpha);

m = sin(alpha) * sin(alpha);
m1 = cos(alpha) * cos(alpha);

%classin = superiorfloat(u,m);
classin='double';
if nargin<3, tol = eps(classin); end


if ~isreal(u) || ~isreal(m)
    error(message('MATLAB:ellipj:ComplexInputs'))
end

if isscalar(m), m = m(ones(size(u))); end
if isscalar(u), u = u(ones(size(m))); end
if ~isequal(size(m),size(u)) 
  error(message('MATLAB:ellipj:InputSizeMismatch')); 
end

mmax = numel(u);

%cn = zeros(size(u),classin);
%cn = zeros(size(u));
cn=u;
sn = cn;dn = sn;

m = m(:).';    % make a row vector
u = u(:).';

if any(m < 0) || any(m > 1), 
  error(message('MATLAB:ellipj:MOutOfRange'));
end

% pre-allocate space and augment if needed
chunk = 10;
%{
a = zeros(chunk,mmax);
c = a;
b = a;
a(1,:) = ones(1,mmax);
c(1,:) = sin(alpha);
b(1,:) = cos(alpha);
%}
c(1) = real(sin(alpha));
b(1) = real(cos(alpha));
a(1) = c(1);
      a = [a; zeros(chunk,mmax)];
      b = [b; zeros(chunk,mmax)];
      c = [c; zeros(chunk,mmax)];
      a(1)=1;

n = zeros(1,mmax);
i = 1;
while any(abs(c(i,:)) > tol) && i<1000
    i = i + 1;
    if i > size(a,1)
      a = [a; zeros(chunk,mmax)];
      b = [b; zeros(chunk,mmax)];
      c = [c; zeros(chunk,mmax)];
    end
    a(i,:) = 0.5 * (a(i-1,:) + b(i-1,:));
    b(i,:) = sqrt(a(i-1,:) .* b(i-1,:));
    c(i,:) = 0.5 * (a(i-1,:) - b(i-1,:));
    in = find((abs(c(i,:)) <= tol) & (abs(c(i-1,:)) > tol));
    if ~isempty(in)
      [mi,ni] = size(in);
      n(in) = repmat((i-1), mi, ni);
    end
end
%phin = zeros(i,mmax,classin);
phin(1)=u;
phin = [phin;zeros(i-1,mmax)];


phin(i,:) = (2 .^ n).*a(i,:).*u;
while i > 1
    i = i - 1;
    in = find(n >= i);
    phin(i,:) = phin(i+1,:);
    if ~isempty(in)
%      phin(i,in) = 0.5 * ...
%      (asin(c(i+1,in).*sin(rem(phin(i+1,in),2*pi))./a(i+1,in)) + phin(i+1,in));
%      phin(i,in) = 0.5 * ...
%      (asin(c(i+1,in).*sin(rem(real(phin(i+1,in)),2*pi))./a(i+1,in)) + phin(i+1,in));
      phin(i,in) = 0.5 * ...
      (asin(c(i+1,in).*sin(real(phin(i+1,in)))./a(i+1,in)) + phin(i+1,in));
    end
end
%sn(:) = sin(rem(real(phin(1,:)),2*pi));
%cn(:) = cos(rem(real(phin(1,:)),2*pi));
sn(:) = sin(real(phin(1,:)));
cn(:) = cos(real(phin(1,:)));
dn(:) = sqrt(1 - m .* (sn(:).').^2);

% special case m = 1 
%m1 = find(m==1);
%sn(m1) = tanh(u(m1));
%cn(m1) = sech(u(m1));
%dn(m1) = sech(u(m1));
% special case m = 0
%dn(m==0) = 1;
