function [Uout,singvals,Vout] = zolosvd(A,a,b,minlen);
%QDWHSVD   Singular value decomposition (SVD) via QDWH and QDWHEIG.
%   [Uout,singvals,Vout] = QDWHSVD(A) computes the SVD of A via 
%   QDWH and QDWHEIG, where A is a rectangular matrix.
%   [Uout,singvals,Vout] = QDWHSVD(A,a,b,minlen) includes optional arguments
%      a: an estimate for nroMA(A,2).
%      b: an estimate for 1/cond(A,1)=\sigma_min(A)/\sigma_max(A).
%      minlen: minimum matrix size for QDWH-eig (default 1).

[m,n] = size(A);
normA = norm(A,'fro');

if m > 1.15*n % Initial QR to reduce to square case.
    [Qini,A] = qr(A,0); m = n;
end

if nargin < 2 || isempty(a)
    a = normest(A,1e-3);
end

if nargin < 3 || isempty(b)
    if m ~= n
        [Q,R] = qr(A,0);
        b = a/condest(R)/1.1;        
    else    
    b = a/condest(A)/1.1;
    end
end

if nargin < 4  || isempty(minlen); minlen = 1; end   
if nargin < 5  || isempty(minlen); minlen = 1; end   

[Uini,HH] = zolopd(A,a,b/a);

if norm(Uini,'fro')^2 < n-0.5
    % Computed Unitary polar factor was not unitary (partial isometry).
    rankdef = 1;
else
    rankdef = 0;
end 

% Done computing Polar decomposition A = Uini*HH, now compute 
% eigendecomposition HH =  Vout*D*Vout'

[Vout,singvals] = zoloeig(HH,normA);

% Accumulate Uini and Vout to get SVD A = Uout*D*Vout 
% (D is not yet computed explicitly).
Vout = 3/2*Vout-Vout*(Vout'*Vout)/2; % Newton-Schulz.
Uout = Uini*Vout;

if exist('Qini'); Uout = Qini*Uout; end % When initial QR used for m>1.15n.
if rankdef == 1
    [Uout R] = qr(Uout,0); % Correction, discussed in paper Sec 5.5.
    Uout = Uout*diag(sign(diag(R)));
end
Uout = 3/2*Uout-Uout*(Uout'*Uout)/2; % Newton-Schulz.

if nargout == 1; Uout = singvals; end